@@ -7,6 +7,7 @@
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/mlvalues.h>
+#include <caml/printexc.h>
 #include <caml/threads.h>
 
 #ifndef Bytes_val
@@ -99,6 +100,7 @@ static void close_av(av_t *av) {
 
   if (av->interrupt_cb != Val_none) {
     caml_remove_generational_global_root(&av->interrupt_cb);
+    av->interrupt_cb = Val_none;
   }
 
   if (av->format_context) {
@@ -158,7 +160,8 @@ AVFormatContext *ocaml_av_get_format_context(value *p_av) {
 
 CAMLprim value ocaml_av_container_options(value unit) {
   CAMLparam0();
-  CAMLreturn((value)avformat_get_class());
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avclass(ret, avformat_get_class()));
 }
 
 CAMLprim value ocaml_av_get_streams(value _av, value _media_type) {
@@ -273,16 +276,15 @@ typedef struct avio_t {
   value seek_cb;
 } avio_t;
 
-#define Avio_val(v) (*(avio_t **)Data_custom_val(v))
-
-static struct custom_operations avio_ops = {
-    "ocaml_avio_ops",    custom_finalize_default,  custom_compare_default,
-    custom_hash_default, custom_serialize_default, custom_deserialize_default};
+#define Avio_val(v) (*(avio_t **)Data_abstract_val(v))
 
 static int ocaml_avio_read_callback(void *private, uint8_t *buf, int buf_size) {
   value buffer, res;
   avio_t *avio = (avio_t *)private;
   int ret;
+  size_t exn_len;
+  char *caml_exn = NULL;
+  char *c_exn = NULL;
 
   ret = caml_c_thread_register();
 
@@ -295,12 +297,32 @@ static int ocaml_avio_read_callback(void *private, uint8_t *buf, int buf_size) {
   res =
       caml_callback3_exn(avio->read_cb, buffer, Val_int(0), Val_int(buf_size));
   if (Is_exception_result(res)) {
+    res = Extract_exception(res);
+
+    caml_exn = caml_format_exception(res);
+    if (caml_exn) {
+      exn_len = strlen(caml_exn) + 1;
+      c_exn = malloc(exn_len);
+      if (!c_exn)
+        caml_raise_out_of_memory();
+      memcpy(c_exn, caml_exn, exn_len);
+      caml_stat_free(caml_exn);
+    }
+
+    if (c_exn) {
+      av_log(avio->avio_context, AV_LOG_ERROR,
+             "Error while executing OCaml read callback: %s\n", c_exn);
+      free(c_exn);
+    }
+
     caml_remove_generational_global_root(&buffer);
     caml_release_runtime_system();
+
     if (ret != 0) {
       caml_c_thread_unregister();
     }
-    return AVERROR_UNKNOWN;
+
+    return AVERROR_EXTERNAL;
   }
 
   if (Int_val(res) < 0) {
@@ -322,6 +344,10 @@ static int ocaml_avio_read_callback(void *private, uint8_t *buf, int buf_size) {
     caml_c_thread_unregister();
   }
 
+  if (Int_val(res) == 0) {
+    return AVERROR_EOF;
+  }
+
   return Int_val(res);
 }
 
@@ -330,6 +356,9 @@ static int ocaml_avio_write_callback(void *private, uint8_t *buf,
   value buffer, res;
   avio_t *avio = (avio_t *)private;
   int ret;
+  size_t exn_len;
+  char *caml_exn = NULL;
+  char *c_exn = NULL;
 
   ret = caml_c_thread_register();
 
@@ -344,12 +373,32 @@ static int ocaml_avio_write_callback(void *private, uint8_t *buf,
   res =
       caml_callback3_exn(avio->write_cb, buffer, Val_int(0), Val_int(buf_size));
   if (Is_exception_result(res)) {
+    res = Extract_exception(res);
+
+    caml_exn = caml_format_exception(res);
+    if (caml_exn) {
+      exn_len = strlen(caml_exn) + 1;
+      c_exn = malloc(exn_len);
+      if (!c_exn)
+        caml_raise_out_of_memory();
+      memcpy(c_exn, caml_exn, exn_len);
+      caml_stat_free(caml_exn);
+    }
+
+    if (c_exn) {
+      av_log(avio->avio_context, AV_LOG_ERROR,
+             "Error while executing OCaml write callback: %s\n", c_exn);
+      free(c_exn);
+    }
+
     caml_remove_generational_global_root(&buffer);
     caml_release_runtime_system();
+
     if (ret != 0) {
       caml_c_thread_unregister();
     }
-    return AVERROR_UNKNOWN;
+
+    return AVERROR_EXTERNAL;
   }
 
   caml_remove_generational_global_root(&buffer);
@@ -406,6 +455,11 @@ static int ocaml_av_interrupt_callback(void *private) {
   av_t *av = (av_t *)private;
   int ret, n;
 
+  // This only happen when closing av_t so
+  // we want to return immediately.
+  if (av->interrupt_cb == Val_none)
+    return 1;
+
   ret = caml_c_thread_register();
 
   caml_acquire_runtime_system();
@@ -496,7 +550,7 @@ CAMLprim value ocaml_av_create_io(value bufsize, value _read_cb,
 
   avio->format_context->pb = avio->avio_context;
 
-  ret = caml_alloc_custom(&avio_ops, sizeof(avio_t *), 0, 1);
+  ret = caml_alloc(1, Abstract_tag);
 
   Avio_val(ret) = avio;
 
@@ -532,16 +586,12 @@ CAMLprim value caml_av_input_io_finalise(value _avio) {
 
 /***** AVInputFormat *****/
 
-static struct custom_operations inputFormat_ops = {
-    "ocaml_av_inputformat",   custom_finalize_default,
-    custom_compare_default,   custom_hash_default,
-    custom_serialize_default, custom_deserialize_default};
-
-void value_of_inputFormat(AVInputFormat *inputFormat, value *p_value) {
+void value_of_inputFormat(avioformat_const AVInputFormat *inputFormat,
+                          value *p_value) {
   if (!inputFormat)
     Fail("Empty input format");
 
-  *p_value = caml_alloc_custom(&inputFormat_ops, sizeof(AVInputFormat *), 0, 1);
+  *p_value = caml_alloc(1, Abstract_tag);
   InputFormat_val((*p_value)) = inputFormat;
 }
 
@@ -555,7 +605,7 @@ CAMLprim value ocaml_av_find_input_format(value _short_name) {
     caml_raise_out_of_memory();
 
   caml_release_runtime_system();
-  AVInputFormat *format = av_find_input_format(short_name);
+  avioformat_const AVInputFormat *format = av_find_input_format(short_name);
   caml_acquire_runtime_system();
 
   free(short_name);
@@ -580,7 +630,7 @@ CAMLprim value ocaml_av_input_format_get_long_name(value _format) {
   CAMLreturn(caml_copy_string(n ? n : ""));
 }
 
-static av_t *open_input(char *url, AVInputFormat *format,
+static av_t *open_input(char *url, avioformat_const AVInputFormat *format,
                         AVFormatContext *format_context, value _interrupt,
                         AVDictionary **options) {
   int err;
@@ -624,6 +674,10 @@ static av_t *open_input(char *url, AVInputFormat *format,
   caml_acquire_runtime_system();
 
   if (err < 0) {
+    if (av->interrupt_cb != Val_none) {
+      caml_remove_generational_global_root(&av->interrupt_cb);
+      av->interrupt_cb = Val_none;
+    }
     free(av);
     if (url)
       free(url);
@@ -637,6 +691,11 @@ static av_t *open_input(char *url, AVInputFormat *format,
   caml_acquire_runtime_system();
 
   if (err < 0) {
+    avformat_free_context(av->format_context);
+    if (av->interrupt_cb != Val_none) {
+      caml_remove_generational_global_root(&av->interrupt_cb);
+      av->interrupt_cb = Val_none;
+    }
     free(av);
     if (url)
       free(url);
@@ -652,7 +711,7 @@ CAMLprim value ocaml_av_open_input(value _url, value _format, value _interrupt,
   CAMLparam4(_url, _format, _interrupt, _opts);
   CAMLlocal3(ret, ans, unused);
   char *url = NULL;
-  AVInputFormat *format = NULL;
+  avioformat_const AVInputFormat *format = NULL;
   int ulen = caml_string_length(_url);
   AVDictionary *options = NULL;
   char *key, *val;
@@ -717,7 +776,7 @@ CAMLprim value ocaml_av_open_input_stream(value _avio, value _format,
   CAMLparam3(_avio, _format, _opts);
   CAMLlocal3(ret, ans, unused);
   avio_t *avio = Avio_val(_avio);
-  AVInputFormat *format = NULL;
+  avioformat_const AVInputFormat *format = NULL;
   AVDictionary *options = NULL;
   char *key, *val;
   int len = Wosize_val(_opts);
@@ -767,7 +826,8 @@ CAMLprim value ocaml_av_open_input_stream(value _avio, value _format,
 
 CAMLprim value ocaml_av_input_obj(value _av) {
   CAMLparam1(_av);
-  CAMLreturn((value)Av_val(_av)->format_context);
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avobj(ret, Av_val(_av)->format_context));
 }
 
 CAMLprim value ocaml_av_get_metadata(value _av, value _stream_index) {
@@ -844,7 +904,8 @@ static stream_t **allocate_input_context(av_t *av) {
   return av->streams;
 }
 
-static stream_t *allocate_stream_context(av_t *av, int index, AVCodec *codec) {
+static stream_t *allocate_stream_context(av_t *av, int index,
+                                         const AVCodec *codec) {
   if (codec) {
     enum AVMediaType type = codec->type;
 
@@ -877,7 +938,7 @@ static stream_t *allocate_stream_context(av_t *av, int index, AVCodec *codec) {
   return stream;
 }
 
-static stream_t *open_stream_index(av_t *av, int index, AVCodec *dec) {
+static stream_t *open_stream_index(av_t *av, int index, const AVCodec *dec) {
   int err;
 
   if (!av->format_context)
@@ -1010,7 +1071,7 @@ CAMLprim value ocaml_av_read_input(value _packet, value _frame, value _av) {
   AVFrame *frame;
   int i, ret, err, frame_kind, skip;
   value _dec;
-  AVCodec *dec = NULL;
+  const AVCodec *dec = NULL;
 
   if (!av->streams && !allocate_input_context(av))
     caml_raise_out_of_memory();
@@ -1058,7 +1119,7 @@ CAMLprim value ocaml_av_read_input(value _packet, value _frame, value _av) {
       }
 
       if (_dec != Val_none) {
-        dec = (AVCodec *)Some_val(_dec);
+        dec = AvCodec_val(Some_val(_dec));
       }
 
       if ((stream = streams[packet->stream_index]) == NULL)
@@ -1234,17 +1295,12 @@ CAMLprim value ocaml_av_seek_bytecode(value *argv, int argn) {
 
 /***** AVOutputFormat *****/
 
-static struct custom_operations outputFormat_ops = {
-    "ocaml_av_outputformat",  custom_finalize_default,
-    custom_compare_default,   custom_hash_default,
-    custom_serialize_default, custom_deserialize_default};
-
-value value_of_outputFormat(AVOutputFormat *outputFormat) {
+value value_of_outputFormat(avioformat_const AVOutputFormat *outputFormat) {
   value v;
   if (!outputFormat)
     Fail("Empty output format");
 
-  v = caml_alloc_custom(&outputFormat_ops, sizeof(AVOutputFormat *), 0, 1);
+  v = caml_alloc(1, Abstract_tag);
   OutputFormat_val(v) = outputFormat;
 
   return v;
@@ -1257,7 +1313,7 @@ CAMLprim value ocaml_av_output_format_guess(value _short_name, value _filename,
   char *short_name = NULL;
   char *filename = NULL;
   char *mime = NULL;
-  AVOutputFormat *guessed;
+  avioformat_const AVOutputFormat *guessed;
 
   if (caml_string_length(_short_name) > 0) {
     short_name =
@@ -1335,9 +1391,9 @@ ocaml_av_output_format_get_subtitle_codec_id(value _output_format) {
       Val_SubtitleCodecID(OutputFormat_val(_output_format)->subtitle_codec));
 }
 
-static av_t *open_output(AVOutputFormat *format, char *file_name,
-                         AVIOContext *avio_context, value _interrupt,
-                         AVDictionary **options) {
+static av_t *open_output(avioformat_const AVOutputFormat *format,
+                         char *file_name, AVIOContext *avio_context,
+                         value _interrupt, AVDictionary **options) {
   int ret;
   AVIOInterruptCB interrupt_cb = {ocaml_av_interrupt_callback, NULL};
   AVIOInterruptCB *interrupt_cb_ptr = NULL;
@@ -1451,7 +1507,7 @@ CAMLprim value ocaml_av_open_output(value _interrupt, value _format,
   CAMLlocal3(ans, ret, unused);
   char *filename =
       strndup(String_val(_filename), caml_string_length(_filename));
-  AVOutputFormat *format = NULL;
+  avioformat_const AVOutputFormat *format = NULL;
   AVDictionary *options = NULL;
   char *key, *val;
   int len = Wosize_val(_opts);
@@ -1518,7 +1574,7 @@ CAMLprim value ocaml_av_open_output_format(value _format, value _opts) {
     }
   }
 
-  AVOutputFormat *format = OutputFormat_val(_format);
+  avioformat_const AVOutputFormat *format = OutputFormat_val(_format);
 
   // open output format
   av_t *av = open_output(format, NULL, NULL, Val_none, &options);
@@ -1552,7 +1608,7 @@ CAMLprim value ocaml_av_open_output_stream(value _format, value _avio,
                                            value _opts) {
   CAMLparam3(_format, _avio, _opts);
   CAMLlocal3(ans, ret, unused);
-  AVOutputFormat *format = OutputFormat_val(_format);
+  avioformat_const AVOutputFormat *format = OutputFormat_val(_format);
   avio_t *avio = Avio_val(_avio);
   AVDictionary *options = NULL;
   char *key, *val;
@@ -1639,7 +1695,7 @@ CAMLprim value ocaml_av_set_metadata(value _av, value _stream_index,
   CAMLreturn(Val_unit);
 }
 
-static stream_t *new_stream(av_t *av, AVCodec *codec) {
+static stream_t *new_stream(av_t *av, const AVCodec *codec) {
   if (!av->format_context)
     Fail("Failed to add stream to closed output");
   if (av->header_written)
@@ -1720,7 +1776,8 @@ static void init_stream_encoder(AVBufferRef *device_ctx, AVBufferRef *frame_ctx,
 }
 
 static stream_t *new_audio_stream(av_t *av, enum AVSampleFormat sample_fmt,
-                                  AVCodec *codec, AVDictionary **options) {
+                                  const AVCodec *codec,
+                                  AVDictionary **options) {
   stream_t *stream = new_stream(av, codec);
 
   AVCodecContext *enc_ctx = stream->codec_context;
@@ -1754,7 +1811,7 @@ CAMLprim value ocaml_av_new_audio_stream(value _av, value _sample_fmt,
                                          value _codec, value _opts) {
   CAMLparam2(_av, _opts);
   CAMLlocal2(ans, unused);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   AVDictionary *options = NULL;
   char *key, *val;
@@ -1798,7 +1855,8 @@ CAMLprim value ocaml_av_new_audio_stream(value _av, value _sample_fmt,
 
 static stream_t *new_video_stream(AVBufferRef *device_ctx,
                                   AVBufferRef *frame_ctx, av_t *av,
-                                  AVCodec *codec, AVDictionary **options) {
+                                  const AVCodec *codec,
+                                  AVDictionary **options) {
   stream_t *stream = new_stream(av, codec);
 
   if (!stream)
@@ -1814,7 +1872,7 @@ CAMLprim value ocaml_av_new_video_stream(value _device_context,
                                          value _codec, value _opts) {
   CAMLparam4(_device_context, _frame_context, _av, _opts);
   CAMLlocal2(ans, unused);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   AVBufferRef *device_ctx = NULL;
   AVBufferRef *frame_ctx = NULL;
@@ -1865,7 +1923,7 @@ CAMLprim value ocaml_av_new_video_stream(value _device_context,
   CAMLreturn(ans);
 }
 
-static stream_t *new_subtitle_stream(av_t *av, AVCodec *codec,
+static stream_t *new_subtitle_stream(av_t *av, const AVCodec *codec,
                                      AVDictionary **options) {
   stream_t *stream = new_stream(av, codec);
   if (!stream)
@@ -1888,7 +1946,7 @@ CAMLprim value ocaml_av_new_subtitle_stream(value _av, value _codec,
                                             value _opts) {
   CAMLparam2(_av, _opts);
   CAMLlocal2(ans, unused);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   AVDictionary *options = NULL;
   char *key, *val;
