@@ -7,6 +7,16 @@ type 'media decoder
 type 'media encoder
 type encode = [ `Encoder ]
 type decode = [ `Decoder ]
+type profile = { id : int; profile_name : string }
+
+type descriptor = {
+  media_type : Avutil.media_type;
+  name : string;
+  long_name : string option;
+  properties : Codec_properties.t list;
+  mime_types : string list;
+  profiles : profile list;
+}
 
 external flag_qscale : unit -> int = "ocaml_avcodec_flag_qscale"
 
@@ -52,6 +62,31 @@ external capabilities : ([< `Audio | `Video ], encode) codec -> capability array
 
 let capabilities c = Array.to_list (capabilities c)
 
+let mk_descriptor d =
+  Option.map
+    (fun (media_type, name, long_name, properties, mime_types, profiles) ->
+      {
+        media_type;
+        name;
+        long_name;
+        properties = Array.to_list properties;
+        mime_types = Array.to_list mime_types;
+        profiles = Array.to_list profiles;
+      })
+    d
+
+external params_descriptor :
+  'media params ->
+  (Avutil.media_type
+  * string
+  * string option
+  * Codec_properties.t array
+  * string array
+  * profile array)
+  option = "ocaml_avcodec_params_descriptor"
+
+let descriptor p = mk_descriptor (params_descriptor p)
+
 type hw_config_method = Hw_config_method.t
 
 type hw_config = {
@@ -187,6 +222,17 @@ module Audio = struct
 
   type id = Codec_id.audio
 
+  external audio_descriptor :
+    id ->
+    (Avutil.media_type
+    * string
+    * string option
+    * Codec_properties.t array
+    * string array
+    * profile array)
+    option = "ocaml_avcodec_audio_descriptor"
+
+  let descriptor id = mk_descriptor (audio_descriptor id)
   let codec_ids = Codec_id.audio
   let get_name = get_name
   let get_description = get_description
@@ -309,6 +355,17 @@ module Video = struct
   type 'mode t = (video, 'mode) codec
   type id = Codec_id.video
 
+  external video_descriptor :
+    id ->
+    (Avutil.media_type
+    * string
+    * string option
+    * Codec_properties.t array
+    * string array
+    * profile array)
+    option = "ocaml_avcodec_video_descriptor"
+
+  let descriptor id = mk_descriptor (video_descriptor id)
   let codec_ids = Codec_id.video
 
   let encoders =
@@ -437,6 +494,17 @@ module Subtitle = struct
   type 'mode t = (subtitle, 'mode) codec
   type id = Codec_id.subtitle
 
+  external subtitle_descriptor :
+    id ->
+    (Avutil.media_type
+    * string
+    * string option
+    * Codec_properties.t array
+    * string array
+    * profile array)
+    option = "ocaml_avcodec_subtitle_descriptor"
+
+  let descriptor id = mk_descriptor (subtitle_descriptor id)
   let codec_ids = Codec_id.subtitle
 
   let encoders =
@@ -526,3 +594,51 @@ let flush_encoder encoder f =
   receive_packet encoder f;
   _flush_encoder encoder;
   receive_packet encoder f
+
+type id = Codec_id.codec_id
+
+external string_of_id : id -> string = "ocaml_avcodec_get_codec_id_name"
+
+module BitstreamFilter = struct
+  type filter = { name : string; codecs : id list; options : Avutil.Options.t }
+  type 'a t
+  type cursor
+
+  external get_next :
+    cursor option -> (string * id array * Avutil.Options.t * cursor) option
+    = "ocaml_avcodec_bsf_next"
+
+  let get_next cursor =
+    Option.map
+      (fun (name, codecs, options, cursor) ->
+        ({ name; codecs = Array.to_list codecs; options }, cursor))
+      (get_next cursor)
+
+  let filters =
+    let rec f cursor filters =
+      match get_next cursor with
+        | None -> filters
+        | Some (filter, cursor) -> f (Some cursor) (filter :: filters)
+    in
+    f None []
+
+  external init :
+    (string * string) array ->
+    string ->
+    'a params ->
+    'a t * 'a params * string array = "ocaml_avcodec_bsf_init"
+
+  let init ?opts { name; _ } params =
+    let opts = opts_default opts in
+    let filter, params, unused = init (mk_opts_array opts) name params in
+    filter_opts unused opts;
+    (filter, params)
+
+  external send_packet : 'a t -> 'a Packet.t -> unit
+    = "ocaml_avcodec_bsf_send_packet"
+
+  external receive_packet : 'a t -> 'a Packet.t
+    = "ocaml_avcodec_bsf_receive_packet"
+
+  external send_eof : 'a t -> unit = "ocaml_avcodec_bsf_send_eof"
+end
