@@ -7,12 +7,19 @@
 #include <caml/mlvalues.h>
 #include <caml/threads.h>
 
-#include "avutil_stubs.h"
-
 #include "avcodec_stubs.h"
+#include "avutil_stubs.h"
 #include "codec_capabilities_stubs.h"
 #include "codec_id_stubs.h"
+#include "codec_properties_stubs.h"
 #include "hw_config_method_stubs.h"
+#include "media_types_stubs.h"
+
+#include "config.h"
+
+#ifdef HAS_BSF_H
+#include <libavcodec/bsf.h>
+#endif
 
 #ifndef AV_PKT_FLAG_DISPOSABLE
 #define AV_PKT_FLAG_DISPOSABLE 0x0010
@@ -58,7 +65,7 @@ CAMLprim value ocaml_avcodec_subtitle_codec_id_to_AVCodecID(value _codec_id) {
 /***** AVCodecContext *****/
 
 static AVCodecContext *create_AVCodecContext(AVCodecParameters *params,
-                                             AVCodec *codec) {
+                                             const AVCodec *codec) {
   AVCodecContext *codec_context;
   int ret = 0;
 
@@ -145,7 +152,12 @@ value value_of_ffmpeg_packet(AVPacket *packet) {
   if (!packet)
     Fail("Empty packet");
 
-  ret = caml_alloc_custom(&packet_ops, sizeof(AVPacket *), 0, 1);
+  int size = 0;
+
+  if (packet->buf)
+    size = packet->buf->size;
+
+  ret = caml_alloc_custom_mem(&packet_ops, sizeof(AVPacket *), size);
   Packet_val(ret) = packet;
 
   return ret;
@@ -339,7 +351,8 @@ static struct custom_operations parser_ops = {
     custom_compare_default,   custom_hash_default,
     custom_serialize_default, custom_deserialize_default};
 
-static parser_t *create_parser(AVCodecParameters *params, AVCodec *codec) {
+static parser_t *create_parser(AVCodecParameters *params,
+                               const AVCodec *codec) {
   parser_t *parser = (parser_t *)calloc(1, sizeof(parser_t));
   if (!parser)
     caml_raise_out_of_memory();
@@ -359,9 +372,9 @@ static parser_t *create_parser(AVCodecParameters *params, AVCodec *codec) {
 }
 
 CAMLprim value ocaml_avcodec_create_parser(value _params, value _codec) {
-  CAMLparam1(_params);
+  CAMLparam2(_params, _codec);
   CAMLlocal1(ans);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
   AVCodecParameters *params = NULL;
 
   if (_params != Val_none)
@@ -434,7 +447,7 @@ CAMLprim value ocaml_avcodec_parse_packet(value _parser, value _data,
 /***** codec_context_t *****/
 
 typedef struct {
-  AVCodec *codec;
+  const AVCodec *codec;
   AVCodecContext *codec_context;
   // output
   int flushed;
@@ -442,7 +455,7 @@ typedef struct {
 
 #define CodecContext_val(v) (*(codec_context_t **)Data_custom_val(v))
 
-static void codec_context(value v) {
+static void finalize_codec_context(value v) {
   codec_context_t *ctx = CodecContext_val(v);
   if (ctx->codec_context)
     avcodec_free_context(&ctx->codec_context);
@@ -451,14 +464,14 @@ static void codec_context(value v) {
 }
 
 static struct custom_operations codec_context_ops = {
-    "ocaml_codec_context",    codec_context,
+    "ocaml_codec_context",    finalize_codec_context,
     custom_compare_default,   custom_hash_default,
     custom_serialize_default, custom_deserialize_default};
 
 CAMLprim value ocaml_avcodec_create_decoder(value _params, value _codec) {
-  CAMLparam1(_params);
+  CAMLparam2(_params, _codec);
   CAMLlocal1(ans);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
   AVCodecParameters *params = NULL;
 
   if (_params != Val_none)
@@ -521,9 +534,9 @@ CAMLprim value ocaml_avcodec_encoder_time_base(value _encoder) {
 
 CAMLprim value ocaml_avcodec_create_audio_encoder(value _sample_fmt,
                                                   value _codec, value _opts) {
-  CAMLparam1(_opts);
+  CAMLparam2(_opts, _codec);
   CAMLlocal3(ret, ans, unused);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   AVDictionary *options = NULL;
   char *key, *val;
@@ -594,7 +607,7 @@ CAMLprim value ocaml_avcodec_create_video_encoder(value _device_context,
                                                   value _opts) {
   CAMLparam3(_device_context, _frame_context, _codec);
   CAMLlocal3(ret, ans, unused);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   AVBufferRef *device_ctx = NULL;
   AVBufferRef *frame_ctx = NULL;
@@ -887,9 +900,10 @@ CAMLprim value ocaml_avcodec_flush_encoder(value _ctx) {
 
 /**** codec ****/
 
-static AVCodec *find_encoder_by_name(const char *name, enum AVMediaType type) {
+static const AVCodec *find_encoder_by_name(const char *name,
+                                           enum AVMediaType type) {
   caml_release_runtime_system();
-  AVCodec *codec = avcodec_find_encoder_by_name(name);
+  const AVCodec *codec = avcodec_find_encoder_by_name(name);
   caml_acquire_runtime_system();
 
   if (!codec || codec->type != type)
@@ -898,9 +912,9 @@ static AVCodec *find_encoder_by_name(const char *name, enum AVMediaType type) {
   return codec;
 }
 
-static AVCodec *find_encoder(enum AVCodecID id, enum AVMediaType type) {
+static const AVCodec *find_encoder(enum AVCodecID id, enum AVMediaType type) {
   caml_release_runtime_system();
-  AVCodec *codec = avcodec_find_encoder(id);
+  const AVCodec *codec = avcodec_find_encoder(id);
   caml_acquire_runtime_system();
 
   if (!codec || codec->type != type)
@@ -909,9 +923,10 @@ static AVCodec *find_encoder(enum AVCodecID id, enum AVMediaType type) {
   return codec;
 }
 
-static AVCodec *find_decoder_by_name(const char *name, enum AVMediaType type) {
+static const AVCodec *find_decoder_by_name(const char *name,
+                                           enum AVMediaType type) {
   caml_release_runtime_system();
-  AVCodec *codec = avcodec_find_decoder_by_name(name);
+  const AVCodec *codec = avcodec_find_decoder_by_name(name);
   caml_acquire_runtime_system();
 
   if (!codec || codec->type != type)
@@ -920,9 +935,9 @@ static AVCodec *find_decoder_by_name(const char *name, enum AVMediaType type) {
   return codec;
 }
 
-static AVCodec *find_decoder(enum AVCodecID id, enum AVMediaType type) {
+static const AVCodec *find_decoder(enum AVCodecID id, enum AVMediaType type) {
   caml_release_runtime_system();
-  AVCodec *codec = avcodec_find_decoder(id);
+  const AVCodec *codec = avcodec_find_decoder(id);
   caml_acquire_runtime_system();
 
   if (!codec || codec->type != type)
@@ -939,20 +954,20 @@ CAMLprim value ocaml_avcodec_parameters_get_bit_rate(value _cp) {
 /**** Audio codec ID ****/
 
 CAMLprim value ocaml_avcodec_get_audio_codec_id(value _codec) {
-  CAMLparam0();
-  AVCodec *codec = (AVCodec *)_codec;
+  CAMLparam1(_codec);
+  const AVCodec *codec = AvCodec_val(_codec);
   CAMLreturn(value_of_audio_codec_id(codec->id));
 }
 
 CAMLprim value ocaml_avcodec_get_video_codec_id(value _codec) {
-  CAMLparam0();
-  AVCodec *codec = (AVCodec *)_codec;
+  CAMLparam1(_codec);
+  const AVCodec *codec = AvCodec_val(_codec);
   CAMLreturn(value_of_video_codec_id(codec->id));
 }
 
 CAMLprim value ocaml_avcodec_get_subtitle_codec_id(value _codec) {
-  CAMLparam0();
-  AVCodec *codec = (AVCodec *)_codec;
+  CAMLparam1(_codec);
+  const AVCodec *codec = AvCodec_val(_codec);
   CAMLreturn(value_of_subtitle_codec_id(codec->id));
 }
 
@@ -962,37 +977,49 @@ CAMLprim value ocaml_avcodec_get_audio_codec_id_name(value _codec_id) {
       avcodec_get_name((enum AVCodecID)AudioCodecID_val(_codec_id))));
 }
 
+CAMLprim value ocaml_avcodec_get_codec_id_name(value _codec_id) {
+  CAMLparam1(_codec_id);
+  CAMLreturn(caml_copy_string(
+      avcodec_get_name((enum AVCodecID)CodecID_val(_codec_id))));
+}
+
 CAMLprim value ocaml_avcodec_find_audio_encoder_by_name(value _name) {
   CAMLparam1(_name);
-  CAMLreturn(
-      (value)find_encoder_by_name(String_val(_name), AVMEDIA_TYPE_AUDIO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_encoder_by_name(String_val(_name), AVMEDIA_TYPE_AUDIO)));
 }
 
 CAMLprim value ocaml_avcodec_find_audio_encoder(value _id) {
   CAMLparam1(_id);
-  CAMLreturn((value)find_encoder(AudioCodecID_val(_id), AVMEDIA_TYPE_AUDIO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_encoder(AudioCodecID_val(_id), AVMEDIA_TYPE_AUDIO)));
 }
 
 CAMLprim value ocaml_avcodec_find_audio_decoder_by_name(value _name) {
   CAMLparam1(_name);
-  CAMLreturn(
-      (value)find_decoder_by_name(String_val(_name), AVMEDIA_TYPE_AUDIO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_decoder_by_name(String_val(_name), AVMEDIA_TYPE_AUDIO)));
 }
 
 CAMLprim value ocaml_avcodec_find_audio_decoder(value _id) {
   CAMLparam1(_id);
-  CAMLreturn((value)find_decoder(AudioCodecID_val(_id), AVMEDIA_TYPE_AUDIO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_decoder(AudioCodecID_val(_id), AVMEDIA_TYPE_AUDIO)));
 }
 
 CAMLprim value ocaml_avcodec_name(value _codec) {
-  CAMLparam0();
-  CAMLreturn(caml_copy_string(((AVCodec *)_codec)->name));
+  CAMLparam1(_codec);
+  CAMLreturn(caml_copy_string(AvCodec_val(_codec)->name));
 }
 
 CAMLprim value ocaml_avcodec_capabilities(value _codec) {
-  CAMLparam0();
+  CAMLparam1(_codec);
   CAMLlocal1(ret);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
   int i, len;
 
   len = 0;
@@ -1010,10 +1037,109 @@ CAMLprim value ocaml_avcodec_capabilities(value _codec) {
   CAMLreturn(ret);
 }
 
-CAMLprim value ocaml_avcodec_hw_methods(value _codec) {
+CAMLprim value ocaml_avcodec_descriptor(enum AVCodecID id) {
   CAMLparam0();
+  CAMLlocal3(ret, tmp, _profile);
+  const AVCodecDescriptor *descriptor = avcodec_descriptor_get(id);
+  int i, len;
+  char **p;
+  struct AVProfile *profile;
+
+  if (!descriptor)
+    CAMLreturn(Val_none);
+
+  ret = caml_alloc_tuple(6);
+
+  Store_field(ret, 0, Val_MediaTypes(descriptor->type));
+  Store_field(ret, 1, caml_copy_string(descriptor->name));
+
+  if (descriptor->long_name) {
+    tmp = caml_alloc_tuple(1);
+    Store_field(tmp, 0, caml_copy_string(descriptor->long_name));
+    Store_field(ret, 2, tmp);
+  } else
+    Store_field(ret, 2, Val_none);
+
+  len = 0;
+  for (i = 0; i < AV_CODEC_PROP_T_TAB_LEN; i++) {
+    if (descriptor->props & AV_CODEC_PROP_T_TAB[i][1])
+      len++;
+  }
+
+  tmp = caml_alloc_tuple(len);
+  len = 0;
+  for (i = 0; i < AV_CODEC_PROP_T_TAB_LEN; i++) {
+    if (descriptor->props & AV_CODEC_PROP_T_TAB[i][1]) {
+      Store_field(tmp, len, AV_CODEC_PROP_T_TAB[i][0]);
+      len++;
+    }
+  }
+  Store_field(ret, 3, tmp);
+
+  len = 0;
+  p = (char **)descriptor->mime_types;
+  while (p && *p) {
+    len++;
+    p++;
+  }
+
+  tmp = caml_alloc_tuple(len);
+  len = 0;
+  p = (char **)descriptor->mime_types;
+  while (p && *p) {
+    Store_field(tmp, len, caml_copy_string(*p));
+    len++;
+    p++;
+  }
+  Store_field(ret, 4, tmp);
+
+  len = 0;
+  profile = (struct AVProfile *)descriptor->profiles;
+  while (profile && profile->profile != FF_PROFILE_UNKNOWN) {
+    len++;
+    profile++;
+  }
+
+  tmp = caml_alloc_tuple(len);
+  len = 0;
+  profile = (struct AVProfile *)descriptor->profiles;
+  while (profile && profile->profile != FF_PROFILE_UNKNOWN) {
+    _profile = caml_alloc_tuple(2);
+    Store_field(_profile, 0, Val_int(profile->profile));
+    Store_field(_profile, 1, caml_copy_string(profile->name));
+    Store_field(tmp, len, _profile);
+    len++;
+    profile++;
+  }
+  Store_field(ret, 5, tmp);
+
+  tmp = caml_alloc_tuple(1);
+  Store_field(tmp, 0, ret);
+
+  CAMLreturn(tmp);
+}
+
+CAMLprim value ocaml_avcodec_params_descriptor(value _params) {
+  CAMLparam1(_params);
+  CAMLreturn(ocaml_avcodec_descriptor(CodecParameters_val(_params)->codec_id));
+}
+
+value ocaml_avcodec_audio_descriptor(value _codec_id) {
+  return ocaml_avcodec_descriptor(AudioCodecID_val(_codec_id));
+}
+
+value ocaml_avcodec_video_descriptor(value _codec_id) {
+  return ocaml_avcodec_descriptor(VideoCodecID_val(_codec_id));
+}
+
+value ocaml_avcodec_subtitle_descriptor(value _codec_id) {
+  return ocaml_avcodec_descriptor(SubtitleCodecID_val(_codec_id));
+}
+
+CAMLprim value ocaml_avcodec_hw_methods(value _codec) {
+  CAMLparam1(_codec);
   CAMLlocal5(ret, tmp1, cons1, tmp2, cons2);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
   int n, i = 0;
   const AVCodecHWConfig *config = avcodec_get_hw_config(codec, i);
 
@@ -1052,11 +1178,11 @@ CAMLprim value ocaml_avcodec_hw_methods(value _codec) {
 }
 
 CAMLprim value ocaml_avcodec_get_supported_channel_layouts(value _codec) {
-  CAMLparam0();
+  CAMLparam1(_codec);
   CAMLlocal2(list, cons);
   int i;
   List_init(list);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   if (codec->channel_layouts) {
     for (i = 0; codec->channel_layouts[i] != -1; i++)
@@ -1067,11 +1193,11 @@ CAMLprim value ocaml_avcodec_get_supported_channel_layouts(value _codec) {
 }
 
 CAMLprim value ocaml_avcodec_get_supported_sample_formats(value _codec) {
-  CAMLparam0();
+  CAMLparam1(_codec);
   CAMLlocal2(list, cons);
   int i;
   List_init(list);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   if (codec->sample_fmts) {
     for (i = 0; codec->sample_fmts[i] != -1; i++)
@@ -1082,11 +1208,11 @@ CAMLprim value ocaml_avcodec_get_supported_sample_formats(value _codec) {
 }
 
 CAMLprim value ocaml_avcodec_get_supported_sample_rates(value _codec) {
-  CAMLparam0();
+  CAMLparam1(_codec);
   CAMLlocal2(list, cons);
   int i;
   List_init(list);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   if (codec->supported_samplerates) {
     for (i = 0; codec->supported_samplerates[i] != 0; i++)
@@ -1159,32 +1285,38 @@ CAMLprim value ocaml_avcodec_get_video_codec_id_name(value _codec_id) {
 
 CAMLprim value ocaml_avcodec_find_video_decoder_by_name(value _name) {
   CAMLparam1(_name);
-  CAMLreturn(
-      (value)find_decoder_by_name(String_val(_name), AVMEDIA_TYPE_VIDEO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_decoder_by_name(String_val(_name), AVMEDIA_TYPE_VIDEO)));
 }
 
 CAMLprim value ocaml_avcodec_find_video_decoder(value _id) {
   CAMLparam1(_id);
-  CAMLreturn((value)find_decoder(VideoCodecID_val(_id), AVMEDIA_TYPE_VIDEO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_decoder(VideoCodecID_val(_id), AVMEDIA_TYPE_VIDEO)));
 }
 
 CAMLprim value ocaml_avcodec_find_video_encoder_by_name(value _name) {
   CAMLparam1(_name);
-  CAMLreturn(
-      (value)find_encoder_by_name(String_val(_name), AVMEDIA_TYPE_VIDEO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_encoder_by_name(String_val(_name), AVMEDIA_TYPE_VIDEO)));
 }
 
 CAMLprim value ocaml_avcodec_find_video_encoder(value _id) {
   CAMLparam1(_id);
-  CAMLreturn((value)find_encoder(VideoCodecID_val(_id), AVMEDIA_TYPE_VIDEO));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_encoder(VideoCodecID_val(_id), AVMEDIA_TYPE_VIDEO)));
 }
 
 CAMLprim value ocaml_avcodec_get_supported_frame_rates(value _codec) {
-  CAMLparam0();
+  CAMLparam1(_codec);
   CAMLlocal3(list, cons, val);
   int i;
   List_init(list);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   if (codec->supported_framerates) {
     for (i = 0; codec->supported_framerates[i].num != 0; i++) {
@@ -1197,11 +1329,11 @@ CAMLprim value ocaml_avcodec_get_supported_frame_rates(value _codec) {
 }
 
 CAMLprim value ocaml_avcodec_get_supported_pixel_formats(value _codec) {
-  CAMLparam0();
+  CAMLparam1(_codec);
   CAMLlocal2(list, cons);
   int i;
   List_init(list);
-  AVCodec *codec = (AVCodec *)_codec;
+  const AVCodec *codec = AvCodec_val(_codec);
 
   if (codec->pix_fmts) {
     for (i = 0; codec->pix_fmts[i] != -1; i++)
@@ -1304,26 +1436,30 @@ CAMLprim value ocaml_avcodec_get_subtitle_codec_id_name(value _codec_id) {
 
 CAMLprim value ocaml_avcodec_find_subtitle_decoder_by_name(value _name) {
   CAMLparam1(_name);
-  CAMLreturn(
-      (value)find_decoder_by_name(String_val(_name), AVMEDIA_TYPE_SUBTITLE));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_decoder_by_name(String_val(_name), AVMEDIA_TYPE_SUBTITLE)));
 }
 
 CAMLprim value ocaml_avcodec_find_subtitle_decoder(value _id) {
   CAMLparam1(_id);
-  CAMLreturn(
-      (value)find_decoder(SubtitleCodecID_val(_id), AVMEDIA_TYPE_SUBTITLE));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_decoder(SubtitleCodecID_val(_id), AVMEDIA_TYPE_SUBTITLE)));
 }
 
 CAMLprim value ocaml_avcodec_find_subtitle_encoder_by_name(value _name) {
   CAMLparam1(_name);
-  CAMLreturn(
-      (value)find_encoder_by_name(String_val(_name), AVMEDIA_TYPE_SUBTITLE));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_encoder_by_name(String_val(_name), AVMEDIA_TYPE_SUBTITLE)));
 }
 
 CAMLprim value ocaml_avcodec_find_subtitle_encoder(value _id) {
   CAMLparam1(_id);
-  CAMLreturn(
-      (value)find_encoder(SubtitleCodecID_val(_id), AVMEDIA_TYPE_SUBTITLE));
+  CAMLlocal1(ret);
+  CAMLreturn(value_of_avcodec(
+      ret, find_encoder(SubtitleCodecID_val(_id), AVMEDIA_TYPE_SUBTITLE)));
 }
 
 /**** Subtitle codec parameters ****/
@@ -1367,7 +1503,7 @@ CAMLprim value ocaml_avcodec_int_of_flag(value _flag) {
 
 CAMLprim value ocaml_avcodec_get_next_codec(value h) {
   CAMLparam0();
-  CAMLlocal4(_id, _h, _ans, _ret);
+  CAMLlocal5(_tmp, _id, _h, _ans, _ret);
   void *s;
   const AVCodec *codec;
   enum AVCodecID id = VALUE_NOT_FOUND;
@@ -1376,7 +1512,7 @@ CAMLprim value ocaml_avcodec_get_next_codec(value h) {
   if (h == Val_int(0)) {
     s = NULL;
   } else {
-    s = (void *)Field(h, 0);
+    s = AvObj_val(Field(h, 0));
   }
 
   codec = av_codec_iterate(&s);
@@ -1408,10 +1544,10 @@ CAMLprim value ocaml_avcodec_get_next_codec(value h) {
   }
 
   _h = caml_alloc_tuple(1);
-  Store_field(_h, 0, (value)s);
+  Store_field(_h, 0, value_of_avobj(_tmp, s));
 
   _ans = caml_alloc_tuple(4);
-  Store_field(_ans, 0, (value)codec);
+  Store_field(_ans, 0, value_of_avcodec(_tmp, codec));
   Store_field(_ans, 1, _id);
   Store_field(_ans, 2, Val_bool(av_codec_is_encoder(codec)));
   Store_field(_ans, 3, _h);
@@ -1424,15 +1560,220 @@ CAMLprim value ocaml_avcodec_get_next_codec(value h) {
 
 CAMLprim value ocaml_avcodec_get_name(value codec) {
   CAMLparam0();
-  CAMLreturn(caml_copy_string(((AVCodec *)codec)->name));
+  CAMLreturn(caml_copy_string(AvCodec_val(codec)->name));
 }
 
 CAMLprim value ocaml_avcodec_get_description(value _codec) {
-  CAMLparam0();
-  AVCodec *codec = (AVCodec *)_codec;
+  CAMLparam1(_codec);
+  const AVCodec *codec = AvCodec_val(_codec);
 
   if (!codec->long_name)
     CAMLreturn(caml_copy_string(""));
 
-  CAMLreturn(caml_copy_string(((AVCodec *)codec)->long_name));
+  CAMLreturn(caml_copy_string(codec->long_name));
+}
+
+#define BsfCursor_val(v) (*(void **)Data_abstract_val(v))
+
+CAMLprim value ocaml_avcodec_bsf_next(value _cursor) {
+  CAMLparam1(_cursor);
+  CAMLlocal2(ans, tmp);
+  int len;
+  enum AVCodecID *codec_id;
+  void *cursor = NULL;
+  if (_cursor != Val_none)
+    cursor = BsfCursor_val(Field(_cursor, 0));
+
+  const AVBitStreamFilter *filter = av_bsf_iterate(&cursor);
+
+  if (!filter)
+    CAMLreturn(Val_none);
+
+  ans = caml_alloc_tuple(4);
+  Store_field(ans, 0, caml_copy_string(filter->name));
+
+  len = 0;
+  codec_id = (enum AVCodecID *)filter->codec_ids;
+  while (codec_id && *codec_id != AV_CODEC_ID_NONE) {
+    codec_id++;
+    len++;
+  }
+
+  tmp = caml_alloc_tuple(len);
+  len = 0;
+  codec_id = (enum AVCodecID *)filter->codec_ids;
+  while (codec_id && *codec_id != AV_CODEC_ID_NONE) {
+    Store_field(tmp, len, Val_CodecID(*codec_id));
+    codec_id++;
+    len++;
+  }
+  Store_field(ans, 1, tmp);
+  Store_field(ans, 2, value_of_avclass(tmp, filter->priv_class));
+
+  tmp = caml_alloc(1, Abstract_tag);
+  BsfCursor_val(tmp) = cursor;
+
+  Store_field(ans, 3, tmp);
+
+  tmp = caml_alloc_tuple(1);
+  Store_field(tmp, 0, ans);
+
+  CAMLreturn(tmp);
+}
+
+#define BsfFilter_val(v) (*(AVBSFContext **)Data_custom_val(v))
+
+static void finalize_bsf_filter(value v) {
+  AVBSFContext *filter = BsfFilter_val(v);
+  av_bsf_free(&filter);
+}
+
+static struct custom_operations bsf_filter_ops = {
+    "bsf_filter_parameters",  finalize_bsf_filter,
+    custom_compare_default,   custom_hash_default,
+    custom_serialize_default, custom_deserialize_default};
+
+CAMLprim value ocaml_avcodec_bsf_init(value _opts, value _name, value _params) {
+  CAMLparam3(_opts, _name, _params);
+  CAMLlocal3(tmp, ans, unused);
+  AVCodecParameters *params = CodecParameters_val(_params);
+  AVBSFContext *bsf;
+  const AVBitStreamFilter *filter;
+  AVDictionary *options = NULL;
+  int ret;
+
+  filter = av_bsf_get_by_name(String_val(_name));
+
+  if (!filter) {
+    caml_raise_not_found();
+  }
+
+  char *key, *val;
+  int len = Wosize_val(_opts);
+  int i, err, count;
+
+  for (i = 0; i < len; i++) {
+    // Dictionaries copy key/values by default!
+    key = (char *)Bytes_val(Field(Field(_opts, i), 0));
+    val = (char *)Bytes_val(Field(Field(_opts, i), 1));
+    err = av_dict_set(&options, key, val, 0);
+    if (err < 0) {
+      av_dict_free(&options);
+      ocaml_avutil_raise_error(err);
+    }
+  }
+
+  caml_release_runtime_system();
+  ret = av_bsf_alloc(filter, &bsf);
+  if (ret < 0) {
+    caml_acquire_runtime_system();
+    ocaml_avutil_raise_error(ret);
+  }
+
+  ret = avcodec_parameters_copy(bsf->par_in, params);
+  if (ret < 0) {
+    av_bsf_free(&bsf);
+    caml_acquire_runtime_system();
+    ocaml_avutil_raise_error(ret);
+  }
+
+  ret = av_opt_set_dict(bsf, &options);
+  if (ret < 0) {
+    av_bsf_free(&bsf);
+    caml_acquire_runtime_system();
+    ocaml_avutil_raise_error(ret);
+  }
+
+  ret = av_bsf_init(bsf);
+  if (ret < 0) {
+    av_bsf_free(&bsf);
+    caml_acquire_runtime_system();
+    ocaml_avutil_raise_error(ret);
+  }
+
+  // Return unused keys
+  count = av_dict_count(options);
+
+  caml_acquire_runtime_system();
+
+  unused = caml_alloc_tuple(count);
+  AVDictionaryEntry *entry = NULL;
+  for (i = 0; i < count; i++) {
+    entry = av_dict_get(options, "", entry, AV_DICT_IGNORE_SUFFIX);
+    Store_field(unused, i, caml_copy_string(entry->key));
+  }
+
+  av_dict_free(&options);
+
+  tmp = caml_alloc_custom(&bsf_filter_ops, sizeof(AVBSFContext *), 0, 1);
+  BsfFilter_val(tmp) = bsf;
+
+  ans = caml_alloc_tuple(3);
+  Store_field(ans, 0, tmp);
+
+  value_of_codec_parameters_copy(bsf->par_out, &tmp);
+  Store_field(ans, 1, tmp);
+
+  Store_field(ans, 2, unused);
+
+  CAMLreturn(ans);
+}
+
+CAMLprim value ocaml_avcodec_bsf_send_packet(value _filter, value _packet) {
+  CAMLparam2(_filter, _packet);
+  int ret;
+  AVPacket *packet = Packet_val(_packet);
+  AVBSFContext *bsf = BsfFilter_val(_filter);
+
+  caml_release_runtime_system();
+  ret = av_bsf_send_packet(bsf, packet);
+  caml_acquire_runtime_system();
+
+  if (ret < 0) {
+    ocaml_avutil_raise_error(ret);
+  }
+
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_avcodec_bsf_send_eof(value _filter) {
+  CAMLparam1(_filter);
+  int ret;
+  AVBSFContext *bsf = BsfFilter_val(_filter);
+
+  caml_release_runtime_system();
+  ret = av_bsf_send_packet(bsf, NULL);
+  caml_acquire_runtime_system();
+
+  if (ret < 0) {
+    ocaml_avutil_raise_error(ret);
+  }
+
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_avcodec_bsf_receive_packet(value _filter) {
+  CAMLparam1(_filter);
+  int ret;
+  AVPacket *packet;
+
+  caml_release_runtime_system();
+  packet = av_packet_alloc();
+
+  if (!packet) {
+    caml_acquire_runtime_system();
+    caml_raise_out_of_memory();
+  }
+
+  ret = av_bsf_receive_packet(BsfFilter_val(_filter), packet);
+
+  if (ret < 0) {
+    av_packet_free(&packet);
+    caml_acquire_runtime_system();
+    ocaml_avutil_raise_error(ret);
+  }
+
+  caml_acquire_runtime_system();
+
+  CAMLreturn(value_of_ffmpeg_packet(packet));
 }
