@@ -13,7 +13,6 @@
 
 #include <libavutil/avassert.h>
 #include <libavutil/avstring.h>
-#include <libavutil/opt.h>
 #include <libavutil/pixdesc.h>
 #include <libavutil/pixfmt.h>
 
@@ -163,55 +162,6 @@ CAMLprim value ocaml_avutil_string_of_error(value error) {
 }
 
 /***** Global initialisation *****/
-#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(58, 9, 100)
-static int lock_manager(void **mtx, enum AVLockOp op) {
-  switch (op) {
-  case AV_LOCK_CREATE:
-    *mtx = malloc(sizeof(pthread_mutex_t));
-
-    if (!*mtx)
-      return 1;
-    return !!pthread_mutex_init(*mtx, NULL);
-
-  case AV_LOCK_OBTAIN:
-    return !!pthread_mutex_lock(*mtx);
-
-  case AV_LOCK_RELEASE:
-    return !!pthread_mutex_unlock(*mtx);
-
-  case AV_LOCK_DESTROY:
-    pthread_mutex_destroy(*mtx);
-    free(*mtx);
-    return 0;
-  }
-  return 1;
-}
-
-CAMLprim value ocaml_avutil_register_lock_manager(value unit) {
-  CAMLparam0();
-  static int registering_done = 0;
-  static pthread_mutex_t registering_mutex = PTHREAD_MUTEX_INITIALIZER;
-
-  if (!registering_done) {
-    pthread_mutex_lock(&registering_mutex);
-
-    if (!registering_done) {
-
-      caml_release_runtime_system();
-      int ret = av_lockmgr_register(lock_manager);
-      caml_acquire_runtime_system();
-
-      if (ret >= 0) {
-        registering_done = 1;
-      }
-      pthread_mutex_unlock(&registering_mutex);
-    }
-  }
-  CAMLreturn(Val_int(registering_done));
-}
-#else
-value ocaml_avutil_register_lock_manager(value unit) { return Val_true; }
-#endif
 
 static pthread_key_t ocaml_c_thread_key;
 static pthread_once_t ocaml_c_thread_key_once = PTHREAD_ONCE_INIT;
@@ -575,14 +525,14 @@ CAMLprim value ocaml_avutil_pixelformat_descriptor(value pixel) {
     Store_field(ret, 6, tmp1);
   } else
     Store_field(ret, 6, Val_none);
-  Store_field(ret, 7, (value)pixdesc);
+  Store_field(ret, 7, value_of_avpixfmtdescriptor(tmp1, pixdesc));
 
   CAMLreturn(ret);
 }
 
 CAMLprim value ocaml_avutil_pixelformat_bits_per_pixel(value d) {
   CAMLparam0();
-  AVPixFmtDescriptor *pixdesc = (AVPixFmtDescriptor *)Field(d, 7);
+  const AVPixFmtDescriptor *pixdesc = AvPixFmtDescriptor_val(Field(d, 7));
 
   CAMLreturn(Val_int(av_get_bits_per_pixel(pixdesc)));
 }
@@ -655,7 +605,14 @@ value value_of_frame(AVFrame *frame) {
   if (!frame)
     Fail("Empty frame");
 
-  ret = caml_alloc_custom(&frame_ops, sizeof(AVFrame *), 0, 1);
+  int size = 0;
+  int n = 0;
+  while (n < AV_NUM_DATA_POINTERS && frame->buf[n] != NULL) {
+    size += frame->buf[n]->size;
+    n++;
+  }
+
+  ret = caml_alloc_custom_mem(&frame_ops, sizeof(AVFrame *), size);
   Frame_val(ret) = frame;
 
   return ret;
@@ -1123,8 +1080,8 @@ CAMLprim value ocaml_avutil_get_opt(value _type, value search_children,
 
   switch (_type) {
   case PVV_String:
-    err = av_opt_get((void *)obj, (const char *)String_val(name), search_flags,
-                     &str);
+    err = av_opt_get(AvObj_val(obj), (const char *)String_val(name),
+                     search_flags, &str);
     if (err < 0)
       ocaml_avutil_raise_error(err);
 
@@ -1250,47 +1207,64 @@ CAMLprim value ocaml_avutil_get_opt(value _type, value search_children,
   }
 }
 
-CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
+CAMLprim value ocaml_avutil_av_opt_iter(value _cursor, value _class) {
   CAMLparam2(_cursor, _class);
   CAMLlocal4(_opt, _type, _tmp, _spec);
+  int unimplement_option = 0;
 
-  const AVClass *_class_cursor;
-  const struct AVOption *_opt_cursor;
+  const AVClass *class;
+  const struct AVOption *option;
+#if LIBAVUTIL_VERSION_INT < AV_VERSION_INT(56, 53, 100)
+  const struct AVOption *cursor;
+#else
+  void *cursor;
+#endif
   AVRational r;
 
   if (_cursor == Val_none) {
-    _opt_cursor = (const struct AVOption *)NULL;
-    _class_cursor = (const AVClass *)_class;
+    cursor = NULL;
+    option = NULL;
+    class = AvClass_val(_class);
   } else {
-    _opt_cursor = (const struct AVOption *)Field(Some_val(_cursor), 0);
-    _class_cursor = (const AVClass *)Field(Some_val(_cursor), 1);
+#if LIBAVUTIL_VERSION_INT < AV_VERSION_INT(56, 53, 100)
+    cursor = AvOptions_val(Field(Some_val(_cursor), 0));
+    option = cursor;
+#else
+    cursor = AvObj_val(Field(Field(Some_val(_cursor), 0), 0));
+    option = AvOptions_val(Field(Field(Some_val(_cursor), 0), 1));
+#endif
+    class = AvClass_val(Field(Some_val(_cursor), 1));
   }
 
-  if (_class_cursor == NULL)
+  if (class == NULL)
     CAMLreturn(Val_none);
 
-  _opt_cursor = av_opt_next(&_class_cursor, _opt_cursor);
+  option = av_opt_next(&class, option);
 
-  if (_opt_cursor == NULL) {
+  if (option == NULL) {
     do {
-      _class_cursor =
-          av_opt_child_class_next((const AVClass *)_class, _class_cursor);
+      class =
+#if LIBAVUTIL_VERSION_INT < AV_VERSION_INT(56, 53, 100)
+          av_opt_child_class_next(AvClass_val(_class), class);
+#else
+          av_opt_child_class_iterate(AvClass_val(_class), &cursor);
+#endif
 
-      if (_class_cursor == NULL)
+      if (class == NULL)
         CAMLreturn(Val_none);
 
-      _opt_cursor = av_opt_next(&_class_cursor, _opt_cursor);
-    } while (_opt_cursor == NULL);
+      option = av_opt_next(&class, option);
+    } while (option == NULL);
   }
 
   _opt = caml_alloc_tuple(6);
-  Store_field(_opt, 0, caml_copy_string(_opt_cursor->name));
+  Store_field(_opt, 0, caml_copy_string(option->name));
 
-  if (_opt_cursor->help == NULL || strlen(_opt_cursor->help) == 0)
+  if (option->help == NULL || strlen(option->help) == 0)
     Store_field(_opt, 1, Val_none);
   else {
     _tmp = caml_alloc_tuple(1);
-    Store_field(_tmp, 0, caml_copy_string(_opt_cursor->help));
+    Store_field(_tmp, 0, caml_copy_string(option->help));
     Store_field(_opt, 1, _tmp);
   }
 
@@ -1300,52 +1274,52 @@ CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
   Store_field(_spec, 1, Val_none);
   Store_field(_spec, 2, Val_none);
 
-  switch (_opt_cursor->type) {
+  switch (option->type) {
   case AV_OPT_TYPE_CONST:
     _type = PVV_Constant;
-    Store_field(_tmp, 0, (value)_opt_cursor);
+    Store_field(_tmp, 0, value_of_avoptions(_cursor, option));
     Store_field(_spec, 0, _tmp);
     break;
   case AV_OPT_TYPE_BOOL:
     _type = PVV_Bool;
-    if (_opt_cursor->default_val.i64 >= 0) {
-      Store_field(_tmp, 0, Val_bool(_opt_cursor->default_val.i64));
+    if (option->default_val.i64 >= 0) {
+      Store_field(_tmp, 0, Val_bool(option->default_val.i64));
       Store_field(_spec, 0, _tmp);
     }
     break;
   case AV_OPT_TYPE_CHANNEL_LAYOUT:
     _type = PVV_Channel_layout;
-    if (av_get_channel_name(_opt_cursor->default_val.i64)) {
-      Store_field(_tmp, 0, Val_ChannelLayout(_opt_cursor->default_val.i64));
+    if (av_get_channel_name(option->default_val.i64)) {
+      Store_field(_tmp, 0, Val_ChannelLayout(option->default_val.i64));
       Store_field(_spec, 0, _tmp);
     }
     break;
   case AV_OPT_TYPE_PIXEL_FMT:
     _type = PVV_Pixel_fmt;
-    if (av_get_pix_fmt_name(_opt_cursor->default_val.i64)) {
-      Store_field(_tmp, 0, Val_PixelFormat(_opt_cursor->default_val.i64));
+    if (av_get_pix_fmt_name(option->default_val.i64)) {
+      Store_field(_tmp, 0, Val_PixelFormat(option->default_val.i64));
       Store_field(_spec, 0, _tmp);
     }
     break;
   case AV_OPT_TYPE_SAMPLE_FMT:
     _type = PVV_Sample_fmt;
-    if (av_get_sample_fmt_name(_opt_cursor->default_val.i64)) {
-      Store_field(_tmp, 0, Val_SampleFormat(_opt_cursor->default_val.i64));
+    if (av_get_sample_fmt_name(option->default_val.i64)) {
+      Store_field(_tmp, 0, Val_SampleFormat(option->default_val.i64));
       Store_field(_spec, 0, _tmp);
     }
     break;
   case AV_OPT_TYPE_INT:
     _type = PVV_Int;
 
-    Store_field(_tmp, 0, Val_int(_opt_cursor->default_val.i64));
+    Store_field(_tmp, 0, Val_int(option->default_val.i64));
     Store_field(_spec, 0, _tmp);
 
     _tmp = caml_alloc_tuple(1);
-    Store_field(_tmp, 0, Val_int(_opt_cursor->min));
+    Store_field(_tmp, 0, Val_int(option->min));
     Store_field(_spec, 1, _tmp);
 
     _tmp = caml_alloc_tuple(1);
-    Store_field(_tmp, 0, Val_int(_opt_cursor->max));
+    Store_field(_tmp, 0, Val_int(option->max));
     Store_field(_spec, 2, _tmp);
     break;
 
@@ -1362,28 +1336,28 @@ CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
     _type = PVV_Duration;
 
   int64_opt:
-    Store_field(_tmp, 0, caml_copy_int64(_opt_cursor->default_val.i64));
+    Store_field(_tmp, 0, caml_copy_int64(option->default_val.i64));
     Store_field(_spec, 0, _tmp);
 
     _tmp = caml_alloc_tuple(1);
 
-    if (_opt_cursor->min <= INT64_MIN)
+    if (option->min <= INT64_MIN)
       Store_field(_tmp, 0, caml_copy_int64(INT64_MIN));
-    else if (_opt_cursor->min >= INT64_MAX)
+    else if (option->min >= INT64_MAX)
       Store_field(_tmp, 0, caml_copy_int64(INT64_MAX));
     else
-      Store_field(_tmp, 0, caml_copy_int64(_opt_cursor->min));
+      Store_field(_tmp, 0, caml_copy_int64(option->min));
 
     Store_field(_spec, 1, _tmp);
 
     _tmp = caml_alloc_tuple(1);
 
-    if (_opt_cursor->max <= INT64_MIN)
+    if (option->max <= INT64_MIN)
       Store_field(_tmp, 0, caml_copy_int64(INT64_MIN));
-    else if (_opt_cursor->max >= INT64_MAX)
+    else if (option->max >= INT64_MAX)
       Store_field(_tmp, 0, caml_copy_int64(INT64_MAX));
     else
-      Store_field(_tmp, 0, caml_copy_int64(_opt_cursor->max));
+      Store_field(_tmp, 0, caml_copy_int64(option->max));
 
     Store_field(_spec, 2, _tmp);
     break;
@@ -1395,15 +1369,15 @@ CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
     _type = PVV_Float;
 
   float_opt:
-    Store_field(_tmp, 0, caml_copy_double(_opt_cursor->default_val.dbl));
+    Store_field(_tmp, 0, caml_copy_double(option->default_val.dbl));
     Store_field(_spec, 0, _tmp);
 
     _tmp = caml_alloc_tuple(1);
-    Store_field(_tmp, 0, caml_copy_double(_opt_cursor->min));
+    Store_field(_tmp, 0, caml_copy_double(option->min));
     Store_field(_spec, 1, _tmp);
 
     _tmp = caml_alloc_tuple(1);
-    Store_field(_tmp, 0, caml_copy_double(_opt_cursor->max));
+    Store_field(_tmp, 0, caml_copy_double(option->max));
     Store_field(_spec, 2, _tmp);
     break;
 
@@ -1411,17 +1385,17 @@ CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
     _type = PVV_Rational;
 
     Store_field(_spec, 0, _tmp);
-    r = av_d2q(_opt_cursor->default_val.dbl, INT_MAX);
+    r = av_d2q(option->default_val.dbl, INT_MAX);
     value_of_rational(&r, &_tmp);
     Store_field(Field(_spec, 0), 0, _tmp);
 
     Store_field(_spec, 1, caml_alloc_tuple(1));
-    r = av_d2q(_opt_cursor->min, INT_MAX);
+    r = av_d2q(option->min, INT_MAX);
     value_of_rational(&r, &_tmp);
     Store_field(Field(_spec, 1), 0, _tmp);
 
     Store_field(_spec, 2, caml_alloc_tuple(1));
-    r = av_d2q(_opt_cursor->max, INT_MAX);
+    r = av_d2q(option->max, INT_MAX);
     value_of_rational(&r, &_tmp);
     Store_field(Field(_spec, 2), 0, _tmp);
     break;
@@ -1445,13 +1419,13 @@ CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
     _type = PVV_String;
 
   string_opt:
-    if (_opt_cursor->default_val.str) {
-      Store_field(_tmp, 0, caml_copy_string(_opt_cursor->default_val.str));
+    if (option->default_val.str) {
+      Store_field(_tmp, 0, caml_copy_string(option->default_val.str));
       Store_field(_spec, 0, _tmp);
     }
     break;
   default:
-    caml_failwith("Invalid option type!");
+    unimplement_option = 1;
   }
 
   _tmp = caml_alloc_tuple(2);
@@ -1459,22 +1433,32 @@ CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
   Store_field(_tmp, 1, _spec);
   Store_field(_opt, 2, _tmp);
 
-  Store_field(_opt, 3, Val_int(_opt_cursor->flags));
+  Store_field(_opt, 3, Val_int(option->flags));
 
-  if (_opt_cursor->unit == NULL || strlen(_opt_cursor->unit) == 0)
+  if (option->unit == NULL || strlen(option->unit) == 0)
     Store_field(_opt, 4, Val_none);
   else {
     _tmp = caml_alloc_tuple(1);
-    Store_field(_tmp, 0, caml_copy_string(_opt_cursor->unit));
+    Store_field(_tmp, 0, caml_copy_string(option->unit));
     Store_field(_opt, 4, _tmp);
   }
 
   _tmp = caml_alloc_tuple(1);
   Store_field(_tmp, 0, caml_alloc_tuple(2));
-  Store_field(Field(_tmp, 0), 0, (value)_opt_cursor);
-  Store_field(Field(_tmp, 0), 1, (value)_class_cursor);
+#if LIBAVUTIL_VERSION_INT < AV_VERSION_INT(56, 53, 100)
+  Store_field(Field(_tmp, 0), 0, value_of_avoptions(_cursor, option));
+#else
+  Store_field(Field(_tmp, 0), 0, caml_alloc_tuple(2));
+  Store_field(Field(Field(_tmp, 0), 0), 0, value_of_avobj(_cursor, cursor));
+  Store_field(Field(Field(_tmp, 0), 0), 1, value_of_avoptions(_cursor, option));
+#endif
+  Store_field(Field(_tmp, 0), 1, value_of_avclass(_class, class));
   Store_field(_opt, 5, _tmp);
 
+  if (unimplement_option)
+    caml_raise_with_arg(*caml_named_value("av_opt_iter_not_implemented"),
+                        Field(_opt, 5));
+
   _tmp = caml_alloc_tuple(1);
   Store_field(_tmp, 0, _opt);
 
@@ -1483,17 +1467,17 @@ CAMLprim value ocaml_avutil_av_opt_next(value _cursor, value _class) {
 
 CAMLprim value ocaml_avutil_avopt_default_int64(value _opt) {
   CAMLparam0();
-  CAMLreturn(caml_copy_int64(((const AVOption *)_opt)->default_val.i64));
+  CAMLreturn(caml_copy_int64(AvOptions_val(_opt)->default_val.i64));
 }
 
 CAMLprim value ocaml_avutil_avopt_default_double(value _opt) {
   CAMLparam0();
-  CAMLreturn(caml_copy_double(((const AVOption *)_opt)->default_val.dbl));
+  CAMLreturn(caml_copy_double(AvOptions_val(_opt)->default_val.dbl));
 }
 
 CAMLprim value ocaml_avutil_avopt_default_string(value _opt) {
   CAMLparam0();
-  CAMLreturn(caml_copy_string(((const AVOption *)_opt)->default_val.str));
+  CAMLreturn(caml_copy_string(AvOptions_val(_opt)->default_val.str));
 }
 
 CAMLprim value ocaml_avutil_av_opt_int_of_flag(value _flag) {
